/*
Google Music Frame :: Frame

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using WebKit;
using Soup;
using Gee;
using Fenryxo;
using Fenryxo.Logger;
using Fenryxo.Config;
using Fenryxo.Path;
using Fenryxo.Widgets;

namespace GMF{
	
	public errordomain FrameError {
		GENERIC
	}
	
	/**
	 * Frame with Google Music web interface.
	 * Handles state changes received from WebView and produced by script
	 * loaded in WebView, console logging, WebView requests, user scripts management, etc.
	 */
	public class Frame: WebView{
		private const string HOME_URL = "http://music.google.com/";
		private const string URL_PREFIX = "http://music.google.com/music/listen";
		private const string URL_PREFIX_SECURE = "https://music.google.com/music/listen";
		private unowned WebFrame info_frame;	
		private string exception;
		private Application app;
		public bool secure{get; private set; default = true;}
		private bool playing = false;
		
		
		/**
		 * Emitted when state of Google Music web interface is changed
		 * 
		 * @param playback_state Playback state ("none", "playing", "paused")
		 * @param can_prev availability of action "go to previous song" 
		 * @param can_next availability of action "go to next song"
		 * @param art_uri URI of album art
		 * @param artist name of the current artist
		 * @param song name of the current song
		 */
		public new signal void data_changed(string playback_state, bool can_prev,
			bool can_next, string? art_uri, string? artist, string? song);
		
		/**
		 * Constructs new Frame for application
		 * 
		 * @param app Application object
		 */
		public class Frame(Application app){
			this.app = app;
			this.secure = this.app.config.bool("secure_connection", true);
			if(this.secure){
				log_message("Secure connection check is enabled.");
			}
			else{
				log_warning("Secure connection check is disabled.");
			}
			this.init_ui();
		}
		
		/**
		 * Initiate GUI
		 */
		private void init_ui(){
			/* Save the current URI to restore previous view on startup */
			this.notify["uri"].connect((o, p) => {
				string uri = (o as WebView).uri;
				log_debug("WebView URI: %s\n", (o as WebView).uri);
				if(uri.has_prefix("https://music.google.com/")
					|| uri.has_prefix("http://music.google.com/")){
					if(this.secure && uri.has_prefix("http://")){
						// Wanna HTTPS
						uri = "https://" + uri.substring(7);
						log_message("Reditect to %s", uri);
						this.open(uri);
					}
					else{
						this.app.ui_state["last_uri"] = (owned) uri;
						try{
							this.app.ui_state.save();
						}
						catch(Error e){
							log_warning("Unable to save UI state: %s", e.message);
						}
						
					}
				}
			});
			
			/* handle debugging */
			this.console_message.connect(this.console_handler);
			
			/* load scripts when Google Music page is loaded */
			this.document_load_finished.connect(this.load_scripts);
			
			/* handle new window request */
			this.new_window_policy_decision_requested.connect(this.on_new_window);
			
			/* handle new window request */
			this.navigation_policy_decision_requested.connect(this.on_new_page);
		}
		
		/**
		 * Loads Google Music home page
		 */
		public void go_home(){
			var uri = this.HOME_URL;
			if(this.secure && uri.has_prefix("http://")){
				uri = "https://" + uri.substring(7);
				log_message("Redirect to %s", uri);
			}
			else if(!this.secure && uri.has_prefix("https://")){
				uri = "http://" + uri.substring(8);
				log_message("Redirect to %s", uri);
			}
			this.open(uri);
		}

		/**
		 * Load Google Music web interface
		 */
		public void start(){
			var uri = this.app.ui_state["last_uri"] ?? this.HOME_URL;
			if(this.secure && uri.has_prefix("http://")){
				uri = "https://" + uri.substring(7);
				log_message("Redirect to %s", uri);
			}
			else if(!this.secure && uri.has_prefix("https://")){
				uri = "http://" + uri.substring(8);
				log_message("Redirect to %s", uri);
			}
			this.show();
			log_debug("Starting Google Music web interface: %s", uri);
			this.open(uri);
		}
		
		private void load_scripts(WebFrame frame){
			log_debug("Loading scripts for frame with uri: %s", frame.get_uri() ?? "null");
			/* check frame and its URI to load scripts only
			 * in Google Music web interface */
			if(frame != this.get_main_frame()
			|| (!this.uri.has_prefix(this.URL_PREFIX)
			 && !this.uri.has_prefix(this.URL_PREFIX_SECURE))) return;
			
			log_debug("Loading the main script");
			var script = get_data_file(this.app.path_name + "/scripts/main.js");
			if(script == null || !script.query_exists()){
				log_critical("Unable to load main script.");
				var err = new ErrorDialog("Error occurred",
				"Main component failed to load.");
				err.run();
				return;
			}
			
			log_debug("Loading script %s", script.get_path());
			string data;
			try{
#if VALAC_0_13
				uint8[] binary_data;
				script.load_contents(null, out binary_data);
				data = (string) binary_data;
				binary_data = null;
#else
				script.load_contents(null, out data);
#endif
			}
			catch(Error e){
				log_critical("Unable to read main script: %s", e.message);
				var err = new ErrorDialog("Error occurred",
				"Main component failed to load.");
				err.show_all();
				err.run();
				return;
			}
			try{
				this.execute(data);
			}
			catch(FrameError e){
				log_critical("Main script execution failed: %s", e.message);
				var err = new ErrorDialog("Error occurred",
				"Main component failed to load.");
				err.run();
				return;
			}
			
			/*
			 * HACK: I haven't found any way how communicate with JavaScript
			 * code in WebView in both directions. So I use
			 * WebView.execute_script() to pass data to WebView,
			 * Frame.title to obtain data from WebView and WebView.console_message()
			 * to provide debug information or exceptions.
			 */
			
			unowned WebFrame info_frame = frame.find_frame("GMF_iframe");
			if(info_frame != null){
				log_debug("Info frame found");
					this.info_frame = info_frame;
					this.info_frame.title_changed.connect(() => {this.update_info();} );
					this.update_info();
			}
			else{
				log_critical("Unable to find info frame");
				var err = new ErrorDialog("Error occurred",
				"Main component failed to load.");
				err.run();
				return;
			}
			
			this.load_user_scripts();
			
		}
		
		/**
		 * Loads user scripts
		 */
		private void load_user_scripts(){
			/* Due to security reasons scripts have to be enabled in configuration */
			string[] scripts = this.app.config.array("user_scripts");
			if(scripts.length == 0){
				log_debug("Frame: No user scripts enabled");
				return;
			}
			
			File[] script_files = list_userscripts();
			foreach(File f in script_files){
				var name = f.get_basename().slice(0, -3);
				if(!(name in scripts)){
					log_debug("User script '%s' is not enabled", name);
					continue;
				}
				try{
					log_debug("Loading user script '%s'", name);
					this.load_user_script(f);
				}
				catch(FrameError e){
					log_warning("User script '%s' failed to load: %s", name, e.message);	
				}
				catch(Error e){
					log_warning("User script '%s' failed to load: %s", name, e.message);	
				}
			}
			
		}
		
		/**
		 * Loads user script
		 * 
		 * @param script script file
		 */
		private void load_user_script(File script) throws Error, FrameError{
			/* load script data - can throw Error */
			string data;
#if VALAC_0_13
				uint8[] binary_data;
				script.load_contents(null, out binary_data);
				data = (string) binary_data;
				binary_data = null;
#else
				script.load_contents(null, out data);
#endif
			/* execute script - can throw FrameError */
			var code = "try{\n\n" + data + "\n\n}catch(e){$throw_exception(e.message);}";
//~ 			log_debug("Code: %s", code);
			this.execute(code);
		}
		
		/**
		 * Process data from WebView
		 */
		private void update_info(){
			/*
			 * Script in WebView stores data as title of info frame (ugly hack).
			 * Items are separated by "§#@!@#§" (ugly again):
			 * 0	playback state
			 * 1	can prev
			 * 2	can next
			 * 3	album art uri
			 * 4	artist
			 * 5	song
			 */
			var title = this.info_frame.title;
			string[] parts = null;
			if(title != null && title != ""){
				parts = this.info_frame.title.split("§#@!@#§");
			}
			
			if(parts != null && parts.length == 3){
				this.playing = parts[0] == "playing";
				this.data_changed(parts[0],
				                  parts[1] == "true",
				                  parts[2] == "true",
				                  null,
				                  null,
				                  null);
			}
			else if(parts != null && parts.length == 6){
				this.playing = parts[0] == "playing";
				this.data_changed(parts[0],
				                  parts[1] == "true",
				                  parts[2] == "true",
				                  parts[3],
				                  parts[4],
				                  parts[5]);
			}
			else{
				this.playing = false;
				this.data_changed("none", false, false, null, null, null);
			}
		}
		
		/**
		 * Execute Google Music player action
		 * 
		 * @param cmd Action
		 * @throw FrameError on failure
		 */
		private void command(string cmd) throws FrameError{
			log_debug("Frame command %s", cmd);
			this.execute("$GMF.command('" + cmd + "');");
		}
		
		/**
		 * Starts playback
		 */
		public void play() throws FrameError{
			if(!this.playing) this.command("playPause");
		}
		
		/**
		 * Starts playback when playing, pauses playback when not playing.
		 */
		public void toggle_play() throws FrameError{
			this.command("playPause");
		}
		
		/**
		 * Pauses playback
		 */
		public void pause() throws FrameError{
			this.command("pauseSong");
		}
		
		/**
		 * Stops playback
		 */
		public void stop() throws FrameError{
			this.pause();
		}
		
		/**
		 * Go to the next song
		 */
		public void prev_song() throws FrameError{
			this.command("prevSong");
		}
		
		/**
		 * Go to the previous song
		 */
		public void next_song() throws FrameError{
			this.command("nextSong");
		}
		
		
		public void thumbs_up() throws FrameError{
			this.execute("$GMF.thumbsUp();");
		}
		
		public void thumbs_down() throws FrameError{
			this.execute("$GMF.thumbsDown();");
		}
		
		/**
		 * JavaScript console wrapper providing log levels and exception passing
		 */
		private bool console_handler(string message, int line_number, string source_id){
			/*
			 * JavaScript provides only console.log() method. GMF use custom prefix
			 * to indicate log level/exception. Exception message is stored
			 * in this.exception which should be checked immediately after calling
			 * this.execute_script(). See this.execute()
			 */
			
			string new_message = null;
			LogLevelFlags level;
			if(message.has_prefix("GMF::debug::")){
				level = LogLevelFlags.LEVEL_DEBUG;
				new_message = message.substring(12);
			}
			else if(message.has_prefix("GMF::info::")){
				level = LogLevelFlags.LEVEL_INFO;
				new_message = message.substring(11);
			}
			else if(message.has_prefix("GMF::warn::")){
				level = LogLevelFlags.LEVEL_WARNING;
				new_message = message.substring(11);
			}
			else if(message.has_prefix("GMF::error::")){
				level = LogLevelFlags.LEVEL_CRITICAL;
				new_message = message.substring(12);
			}
			else if(message.has_prefix("GMF::exception::")){
				this.exception = message.substring(16);
				return true;
			}
			else{
				level = LogLevelFlags.LEVEL_CRITICAL;
			}
			
			log_with_level(level, "%s %d: %s",
				source_id, line_number, new_message ?? message);
			return true;
		}
		
		/**
		 * Executes script and propagates exceptions
		 */
		private void execute(string cmd) throws FrameError{
			this.execute_script(cmd);
			/* Exceptions are passed via Javascript console, see this.console_handler() */
			if(this.exception != null){
				var msg = (owned) this.exception;
				this.exception = null;
				throw new FrameError.GENERIC(msg);
			}
		}
		
		/**
		 * Open new window in the default web browser
		 */
		private bool on_new_window(WebFrame frame, NetworkRequest req,
			WebNavigationAction action, WebPolicyDecision decision){
			unowned string uri = req.get_uri();
			log_debug("New window request: %s", uri);
			
			if(("accounts.google.com/" in uri)
				|| ("google.com/accounts" in uri)){
				this.open(uri);
			}
			else{
				this.open_in_web_browser(uri);
			}
			decision.ignore();
			return true;
		}
		
		/**
		 * Decide whether open page in frame or in the default browser
		 */
		private bool on_new_page(WebFrame frame, NetworkRequest req,
			WebNavigationAction action, WebPolicyDecision decision){
			unowned string uri = req.get_uri();
			log_debug("New page request: URI: %s, frame URI: %s",
				uri, frame.get_uri() ?? "null");
			
			if(frame != this.get_main_frame()){
				// don't care about pages in <iframe> elements
				return false;
			}
			
			/* Open Flash install page in the default browser */
			if(uri.has_prefix("http://get.adobe.com/flashplayer")){
				this.open_in_web_browser(uri);
				decision.ignore();
				return true;
			}
			
			/* Open Google services in the default browser (except login page) */
			if(("google.com" in uri)
				&& !("music.google.com" in uri)
				&& !("accounts.google.com/" in uri)
				&& !("google.com/accounts" in uri)){
				this.open_in_web_browser(uri);
				decision.ignore();
				return true;
			}
			
			return false;
		}
		
		/**
		 * Opens page in the default web browser
		 * 
		 * @param uri URI to open
		 */
		private void open_in_web_browser(string uri){
			try{
				Fenryxo.System.open_uri(uri);
			}
			catch(SpawnError e){
				log_critical("Unable to open web browser for %s. Cause: %s", uri, e.message);
				var err = new ErrorDialog("Error occured",
				@"Failed to open $uri in your default web browser.");
				err.run();
			}
		}
	}
	
	/**
	 * Provides list of user script files
	 */
	private static File[] list_userscripts(){
		File[] result = {};
		string[] names = {}; // base names without extension
		FileInfo file_info;
		File script_file;
		
		foreach(weak File dir in get_data_dirs(GMF.Config.APPNAME + "/userscripts")){
			try{
				var enumerator = dir.enumerate_children(FILE_ATTRIBUTE_STANDARD_NAME, 0);
				while((file_info = enumerator.next_file()) != null){
					weak string name = file_info.get_name();
					// JavaScript extension requred
					if(! name.has_suffix(".js")) continue;
					// Skip if script has been already found in one of previous dirs
					if(name in names) continue;
					script_file = dir.get_child(file_info.get_name());
					if(script_file.query_file_type(0) != FileType.REGULAR) continue;
					names += name;
					result += script_file;
				}
			}
			catch(Error e){
				continue; // Not a directory?
			} 
		}
		return result;
	}	
}
